#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <malloc.h>
#include <assert.h>

#include "Matrici.h"

CODICEUSCITA readMatrice(Matrice m, int righe, int colonne)
{
    char formato[100] = "";
    int riga, colonna;
    for (colonna = 0; colonna < colonne; colonna++)
        strcat(formato, ELEMENTFORMAT);
    for (riga = 0; riga < righe; riga++)
    {
        printf("Inserire la riga %d: ", riga);
        if (scanf(formato, m[riga]) != colonne)
            return VALORINONVALIDI;
    }
    return OK;
}

void printMatrice(Matrice m, int righe, int colonne)
{
    int riga, colonna;
    char formato[10] = "";
    strcat(formato, ELEMENTFORMAT);
    strcat(formato, "\t"); //Minimizzare il numero di spazi!
    for (riga = 0; riga < righe; riga++)
    {
        for (colonna = 0; colonna < colonne; colonna++)
            printf(formato, m[riga][colonna]);
        printf("\n");
    }
    printf("\n");
}

void trasposta(Matrice m, Matrice tm, int dim)
{
    int riga, colonna;
    for (riga = 0; riga < dim; riga++)
        for (colonna = 0; colonna < dim; colonna++)
            tm[colonna][riga] = m[riga][colonna];
}

void estraiMinore(Matrice minore, Matrice m, int dim, int rigaElemento, int colonnaElemento)
{
    int riga, colonna;
    for (riga = 0; riga < dim; riga++)
    {
        int rigaMinore  = (riga < rigaElemento) ? riga : riga - 1;
        for (colonna = 0; colonna < dim; colonna++)
        {
            if (riga != rigaElemento && colonna != colonnaElemento)
            {
                int colonnaMinore = colonna < colonnaElemento ? colonna : colonna - 1;
                minore[rigaMinore][colonnaMinore] = m[riga][colonna];
            }
        }
    }
}

ELEMENT determinante(Matrice m, int dim)
{
    int riga = 0, colonna;
    if (dim == 2)
    {
        return m[0][0] * m[1][1] - m[0][1] * m[1][0];
    }
    else
    {
        ELEMENT det = 0;
        for (colonna = 0; colonna < dim; colonna++)
        {
            Matrice minore;
            ELEMENT detMinore;
            estraiMinore(minore, m, dim, riga, colonna);
            printMatrice(minore, dim - 1, dim - 1);
            detMinore = determinante(minore, dim - 1);
            det += m[riga][colonna] * (colonna % 2 == 0 ? detMinore : -detMinore);
        }
        return det;
    }
}

void triang(Matrice a, int rows, int columns)
{
    int masterEq, coeff, currentEq;
    for (masterEq = 0; masterEq < rows; masterEq++)
    {
        for (currentEq = masterEq + 1; currentEq < rows; currentEq++)
        {
            ELEMENT m = a[currentEq][masterEq]/a[masterEq][masterEq];
            for (coeff = masterEq; coeff < columns; coeff++)
                a[currentEq][coeff] = a[currentEq][coeff] - m * a[masterEq][coeff];
        }
    }
}

void elim_indietro(ELEMENT *x, Matrice a, int rows)
{ 
    int currentEq, coeff;
    for (currentEq = rows - 1; currentEq >= 0; currentEq--)
    {   
        for (x[currentEq] = 0, coeff = currentEq + 1; coeff < rows; coeff++)
            x[currentEq] -= a[currentEq][coeff] * x[coeff];
        x[currentEq] += a[currentEq][rows];
        x[currentEq] /= a[currentEq][currentEq];
    }
}


void testDeterminante()
{
    Matrice m1 = { {1, 1, 1, 1}, {0, 1, 1, 1}, {0, 0, 1, 1}, {0, 0, 0, 1} };
    printf("%f", determinante(m1, 4));
}

void testGauss()
{
    Matrice a = {{1, 2, 3, 4}, {5, 6, 7, 10}, {7, 11, 13, 17}};
    ELEMENT x[MAXROWS];
    int i = 0;
    
    triang(a, 3, 4);
    printMatrice(a, 3, 4);
    elim_indietro(x, a, 3);
    for (; i < 3; i++)
        printf("x%d = %f\n", i + 1, x[i]);
}

/********************************************************************/
/* DYN MATRIX */
/********************************************************************/

DynMatrix newDynMatrix(int rows, int cols)
{
    DynMatrix dm = (DynMatrix)malloc(sizeof(DynMatrix));
    if (dm == NULL)
        return NULL;
    dm->matrix = (ELEMENT*)malloc(sizeof(ELEMENT) * rows * cols);
    if (dm->matrix == NULL)
        return NULL;
    dm->rows = rows;
    dm->cols = cols;
    return dm;
}

void destroyDynMatrix(DynMatrix m)
{
    free(m->matrix);
    free(m);
}

BOOLEAN checkBounds(DynMatrix m, int row, int col)
{
    return row >= 0 && col >= 0 && row < m->rows && col < m->cols;
}

ELEMENT getMatrixElement(DynMatrix m, int row, int col)
{
    assert(checkBounds(m, row, col));
    return *(m->matrix + row * m->cols + col);
}

void setMatrixElement(DynMatrix m, int row, int col, ELEMENT value)
{
    assert(checkBounds(m, row, col));
    *(m->matrix + row * m->cols + col) = value;
}

int getRowCount(DynMatrix m)
{
    return m->rows;
}

int getColCount(DynMatrix m)
{
    return m->cols;
}

void printMatrice_dyn(DynMatrix m)
{
    int row, col;
    char format[10] = "";
    strcat(format, ELEMENTFORMAT);
    strcat(format, "\t"); //Minimizzare il numero di spazi!
    for (row = 0; row < getRowCount(m); row++)
    {
        for (col = 0; col < getColCount(m); col++)
            printf(format, getMatrixElement(m, row, col));
        printf("\n");
    }
    printf("\n");
}

DynMatrix estraiMinore_dyn(DynMatrix m, int elRow, int elColumn)
{
    int row, col;
    DynMatrix minore;
    assert(getRowCount(m) == getColCount(m)); //matrice quadrata
    assert(checkBounds(m, elRow, elColumn));
    minore = newDynMatrix(getRowCount(m) - 1, getColCount(m) - 1);
    for (row = 0; row < getRowCount(m); row++)
    {
        int rowMinor  = (row < elRow) ? row : row - 1;
        for (col = 0; col < getColCount(m); col++)
        {
            if (row != elRow && col != elColumn)
            {
                int colMinor = col < elColumn ? col : col - 1;
                setMatrixElement(minore, rowMinor, colMinor, getMatrixElement(m, row, col));
            }
        }
    }
    return minore;
}

ELEMENT determinante_dyn(DynMatrix m)
{
    int riga = 0, colonna;
    assert(getRowCount(m) == getColCount(m)); //matrice quadrata
    if (getRowCount(m) == 2)
    {
        return getMatrixElement(m, 0, 0) * getMatrixElement(m, 1, 1) -
            getMatrixElement(m, 0, 1) * getMatrixElement(m, 1, 0);
    }
    else
    {
        ELEMENT det = 0;
        for (colonna = 0; colonna < getColCount(m); colonna++)
        {
            DynMatrix minore;
            ELEMENT detMinore;
            minore = estraiMinore_dyn(m, riga, colonna);
            printMatrice_dyn(minore);
            detMinore = determinante_dyn(minore);
            det += getMatrixElement(m, riga, colonna) * (colonna % 2 == 0 ? detMinore : -detMinore);
            destroyDynMatrix(minore);
        }
        return det;
    }
}

void testDeterminante_dyn()
{
    Matrice m1 = { {1, 1, 1, 1}, {0, 1, 1, 1}, {0, 0, 1, 1}, {0, 0, 0, 1} };
    DynMatrix m = newDynMatrix(4, 4);
    m->matrix = (ELEMENT*)m1;
    printf("%f", determinante_dyn(m));
    destroyDynMatrix(m);
}

